#include "toposens_pointcloud/mapping.h"
#include <toposens_pointcloud/logging.h>

namespace toposens_pointcloud
{
/** Each incoming scan is converted to a new PointCloud message of
 *  template XYZI, which corresponds well with a TsPoint structure.
 *  A second persistent PointCloud structure holds the cumulative
 *  of all scans received during a run. At the end of each run, this
 *  cumulative data is saved as a PCD file in the package folder for
 *  subsequent use with other PCL tools (like pcl_octree_viewer).
 *
 *  Prior to pointcloud conversion, each scan can also be optionally
 *  transformed to a different reference frame (usually, 'world' or
 *  'base_link' frame). This can be used, for instance, to efficiently
 *  map an area using sensor or odometry data.
 */
Mapping::Mapping(ros::NodeHandle nh, ros::NodeHandle private_nh)
{
  _tf2_listener = new tf2_ros::TransformListener(_tf2_buffer);

  private_nh.param<std::string>("scans_topic", _scans_topic, "ts_scans");

  ROS_INFO("Reading scans from \"%s\"", _scans_topic.c_str());

  _scans_sub = nh.subscribe(_scans_topic, 100, &Mapping::_convert, this);
  _cloud_pub = nh.advertise<XYZINCloud>(kPointCloudTopic, 100);

  private_nh.param<std::string>("target_frame", _target_frame, "toposens");
  private_nh.param<float>("lifetime_normals_vis", _lifetime_normals_vis, 0.0);

  if (_lifetime_normals_vis != 0.0)
    _marker_pub = nh.advertise<visualization_msgs::Marker>(kPointCloudTopic + "_normals", 100);

  int pcd_save_interval;
  private_nh.param<int>("pcd_save_interval", pcd_save_interval, 1);
  if(pcd_save_interval>0) _log_thread = new boost::thread(&Logging::logPointcloud);
}


/** Destructor frees memory from heap */
Mapping::~Mapping()
{
  delete _tf2_listener;

  if (_log_thread != nullptr)
  {
    _log_thread->interrupt();
    _log_thread->join();
    delete _log_thread;
  }
}


/** Waits for a target frame transform to become available and maps
 *  incoming TsPoints to this frame. Transformed points are added
 *  to a instantaneous pointcloud structure for publishing.
 */
void Mapping::_convert(const toposens_msgs::TsScan::ConstPtr& msg)
{
  XYZINCloud::Ptr tc(new XYZINCloud);
  pcl_conversions::toPCL(msg->header.stamp, tc->header.stamp);
  tc->header.frame_id = _target_frame;
  tc->height = 1;

  geometry_msgs::TransformStamped transform;

  try
  {
    transform = _tf2_buffer.lookupTransform(_target_frame,
                                            msg->header.frame_id,
                                            msg->header.stamp,
                                            ros::Duration(0.5));

    for (auto it = msg->points.begin(); it != msg->points.end(); ++it)
    {
      if (it->intensity <= 0) continue;
      toposens_msgs::TsPoint pt = *it;
      pcl::PointXYZINormal p = _convertToXYZINormal(pt);
      _transform(transform, p, msg->header);
      tc->points.push_back(p);
    }
  }
  catch (tf2::TransformException ex)
  {
    ROS_INFO_STREAM(ex.what());
  }

  tc->width = tc->points.size();
  _cloud_pub.publish(tc);

  if(_lifetime_normals_vis != 0.0) _publishNormals(tc);
}


/** Converts a PointXYZINormal to intermediary PointStamped and
 *  Vector3Stamped messages compatible with the transform listener.
 *  After transformation converts them back to a PointXYZINormal.
 *  Transformation is skipped if #target_frame is same as the point's
 *  current frame.
 */
void Mapping::_transform(geometry_msgs::TransformStamped transform,
                         pcl::PointXYZINormal& pt,
                         std_msgs::Header h)
{
  // Skip if current and target frames are the same
  if (h.frame_id == _target_frame) return;

  // Intermediate stamped point for transformation
  geometry_msgs::PointStamped ps;
  ps.point.x = pt.x;
  ps.point.y = pt.y;
  ps.point.z = pt.z;
  ps.header = h;

  // Intermediate stamped vector for transformation
  geometry_msgs::Vector3Stamped normal;
  normal.vector.x = pt.normal_x;
  normal.vector.y = pt.normal_y;
  normal.vector.z = pt.normal_z;
  normal.header = h;

  // Perform the transformations to target frames
  try
  {
    tf2::doTransform(ps, ps, transform);
    tf2::doTransform(normal, normal, transform);
  }
  catch (tf2::TransformException ex)
  {
    ROS_ERROR("Point or normal transformation failed: %s", ex.what());
  }

  // Update the input PCL point
  pt.x = ps.point.x;
  pt.y = ps.point.y;
  pt.z = ps.point.z;
  pt.normal_x = normal.vector.x;
  pt.normal_y = normal.vector.y;
  pt.normal_z = normal.vector.z;
}


/** Transforms a TsPoint to a PointXYZINormal.
 *  The normal is generated as a vector pointing from the detected point
 *  to the sensor origin with unit length.
 */
pcl::PointXYZINormal Mapping::_convertToXYZINormal(toposens_msgs::TsPoint i)
{
  // Fast calculation of inverse of square roots
  // https://en.wikipedia.org/wiki/Fast_inverse_square_root
  auto isqrtf = [](float number)
  {
    long i;
    float x2, y;
    const float threehalfs = 1.5F;

    x2 = number * 0.5F;
    y  = number;
    i  = * ( long * ) &y;
    i  = 0x5f3759df - ( i >> 1 );
    y  = * ( float * ) &i;
    y  = y * ( threehalfs - ( x2 * y * y ) );
    return y;
  };

  pcl::PointXYZINormal o;

  o.x = i.location.x;
  o.y = i.location.y;
  o.z = i.location.z;
  o.intensity = i.intensity;

  // Approximation of the inverse of the length of the normal vector
  float len = isqrtf(i.location.x * i.location.x +
                     i.location.y * i.location.y +
                     i.location.z * i.location.z);
  // Normalize and store the vector in the PCL point
  o.normal_x = -i.location.x * len;
  o.normal_y = -i.location.y * len;
  o.normal_z = -i.location.z * len;

  return o;
}


/** Publishes markers that contain arrows that visualize the normal vectors of
 *  points. Each arrow is defined by a starting point and an end point.
 */
void Mapping::_publishNormals(XYZINCloud::ConstPtr tc)
{
  // Marker for the arrow(s) to be published
  visualization_msgs::Marker marker;

  // Initialize the marker
  marker.type = visualization_msgs::Marker::ARROW;
  marker.header = pcl_conversions::fromPCL(tc->header);
  marker.ns = "normals";  // Namespace the normal marker will be posted to
  marker.action = visualization_msgs::Marker::ADD;
  marker.lifetime = ros::Duration(_lifetime_normals_vis);
  marker.pose.orientation.x = 0.0; marker.pose.orientation.y = 0.0;
  marker.pose.orientation.z = 0.0; marker.pose.orientation.w = 1.0;
  marker.scale.x = 0.005f; marker.scale.y = 0.01f;
  marker.color.r = 1.f; marker.color.g = 0.f;
  marker.color.b = 0.f; marker.color.a = 1.f;

  // Publish normal markers individually
  for (auto p : tc->points)
  {
    marker.id = _marker_id++;

    marker.points.clear();
    geometry_msgs::Point pt;

    pt.x = p.x;
    pt.y = p.y;
    pt.z = p.z;

    marker.points.push_back(pt);

    pt.x += p.normal_x * .1f;
    pt.y += p.normal_y * .1f;
    pt.z += p.normal_z * .1f;

    marker.points.push_back(pt);
    _marker_pub.publish(marker);
  }
}

} // namespace toposens_pointcloud
